/* $Id: create_table.c,v 1.5 2012/01/15 17:40:09 jim Exp $
 *
 * This file is part of the VIRCAM Pipeline
 * Copyright (C) 2005 Cambridge Astronomy Survey Unit
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: jim $
 * $Date: 2012/01/15 17:40:09 $
 * $Revision: 1.5 $
 * $Name: vcam-1_3_2 $
 */

#include <stdio.h>
#include <string.h>
#include "imcore.h"

static float *work = NULL;

static void tidy(void);

/**@{*/

/*---------------------------------------------------------------------------*/
/**
    \ingroup cataloguemodules
    \brief Initialise catalogues
  
    \par Name:
        tabinit
    \par Purpose:
        Initialise the output table.
    \par Description:
        Wrapper routine to call the relevant initialisation routine for
	each of the allowed types of catalogues.
    \par Language:
        C
    \param ap
        The current ap structure
    \returns
        Nothing
    \par QC headers:
        None
    \par DRS headers:
        None
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

extern void tabinit(ap_t *ap) {

    switch (cattype) {
    case CAT_INTWFC:
	tabinit_1();
	break;
    case CAT_WFCAM:
	tabinit_2();
	break;
    case CAT_BASIC:
	tabinit_3();
	break;
    case CAT_OBJMASK:
	tabinit_4(ap);
	break;
    default:
	cpl_msg_error("tabinit","Option %" CPL_SIZE_FORMAT " does not exist",
		      (cpl_size)cattype);
	tab = NULL;
	break;
    }
}

/*---------------------------------------------------------------------------*/
/**
    \ingroup cataloguemodules
    \brief Do seeing estimate
  
    \par Name:
        do_seeing
    \par Purpose:
        Do the seeing estimate 
    \par Description:
        Wrapper routine to call the relevant routine to work out the seeing
	for each of the allowed types of catalogues
    \par Language:
        C
    \param ap
        The current ap structure
    \retval VIR_OK
        If all went well
    \retval VIR_FATAL
        If catalogue type is unrecognised
    \par QC headers:
        None
    \par DRS headers:
        None
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

extern int do_seeing(ap_t *ap) {
    int status;

    switch (cattype) {
    case CAT_INTWFC:
	status = do_seeing_1(ap);
	break;
    case CAT_WFCAM:
	status = do_seeing_2(ap);
	break;
    case CAT_BASIC:
	status = do_seeing_3(ap);
	break;
    case CAT_OBJMASK:
	status = do_seeing_4(ap);
	break;
    default:
	status = VIR_FATAL;
	cpl_msg_error("do_seeing","Option %" CPL_SIZE_FORMAT " does not exist",
		      (cpl_size)cattype);
	break;
    }
    return(status);
}

/*---------------------------------------------------------------------------*/
/**
    \ingroup cataloguemodules
    \brief Process results
  
    \par Name:
        process_results
    \par Purpose:
        Process the results for each object and store them in the table
    \par Description:
        Wrapper routine to call the relevant routine to work out the results
	for each of the allowed types of catalogues
    \par Language:
        C
    \param ap
        The current ap structure
    \retval VIR_OK
        If all went well
    \retval VIR_FATAL
        If catalogue type is unrecognised
    \par QC headers:
        None
    \par DRS headers:
        None
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

extern int process_results(ap_t *ap) {
    int status;
    
    switch (cattype) {
    case CAT_INTWFC:
	status = process_results_1(ap);
	break;
    case CAT_WFCAM:
	status = process_results_2(ap);
	break;
    case CAT_BASIC:
	status = process_results_3(ap);
	break;
    case CAT_OBJMASK:
	status = process_results_4(ap);
	break;
    default:
	status = VIR_FATAL;
	cpl_msg_error("process_result","Option %" CPL_SIZE_FORMAT " does not exist",
		      (cpl_size)cattype);
	break;
    }
    return(status);
}

/*---------------------------------------------------------------------------*/
/**
    \ingroup cataloguemodules
    \brief Close the table structure
  
    \par Name:
        tabclose
    \par Purpose:
        Close the table structure
    \par Description:
        Wrapper routine to call the relevant routine to close the table
	for each of the allowed types of catalogues
    \par Language:
        C
    \param ap
        The current ap structure
    \retval VIR_OK
        If all went well
    \retval VIR_FATAL
        If catalogue type is unrecognised
    \par QC headers:
        None
    \par DRS headers:
        None
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

extern int tabclose(ap_t *ap) {
    int status;

    switch (cattype) {
    case CAT_OBJMASK:
	status = tabclose_4(ap);
        break;
    default:
	status = VIR_OK;
	break;
    }
    return(status);
}

/*---------------------------------------------------------------------------*/
/**
    \ingroup cataloguemodules
    \brief Initialise tables (generic)
  
    \par Name:
        tabinit_gen
    \par Purpose:
        Initialise tables (generic)
    \par Description:
        Generic routine to create FITS tables for the output catalogues
    \par Language:
        C
    \param ncols
        The number of columns in the table
    \param ttype
        Array of column names for FITS table
    \param tunit
        Array of units for each of the columns
    \param tform
        Array of formats for each of the columns as defined in the FITS
	standard
    \returns
        Nothing
    \par QC headers:
        None
    \par DRS headers:
        None
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

extern void tabinit_gen(int ncols, const char *ttype[], const char *tunit[], 
			cpl_type tform[]) {
    int i;
    const char *fctid = "tabinit_gen";
    
    /* First, create the table with a default number of rows. */

    if ((tab = cpl_table_new(0)) == NULL) {
        cpl_msg_error(fctid,"Unable to open cpl table!");
	return;
    }

    /* Now define all of the columns */

    for (i = 0; i < ncols; i++) {
	cpl_table_new_column(tab,ttype[i],tform[i]);
	cpl_table_set_column_unit(tab,ttype[i],tunit[i]);
    }

}

/*---------------------------------------------------------------------------*/
/**
    \ingroup cataloguemodules
    \brief Do seeing estimate (generic)
  
    \par Name:
        do_seeing_gen
    \par Purpose:
        Do seeing estimate (generic)
    \par Description:
        Wrapper routine for doing the seeing estimate
    \par Language:
        C
    \param ap
        The current ap structure
    \param col_ellipt
        The name of the column for ellipticity
    \param col_pkht
        The name of the column for the peak height
    \param col_areals
        The array of names of the areal profile columns
    \retval VIR_OK
        If all is ok. This is currently the only value.
    \par QC headers:
        None
    \par DRS headers:
        None
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

extern int do_seeing_gen(ap_t *ap, const char *col_ellipt, 
			 const char *col_pkht, char *col_areals[NAREAL]) {
    int i;
    float fwhm,*areal[NAREAL],*ellipt,*pkht;

    /* Get some space and read the relevant columns */

    if (nobjects >= 3) {
	ellipt = cpl_table_get_data_float(tab,col_ellipt);
	pkht = cpl_table_get_data_float(tab,col_pkht);
	work = cpl_malloc(nobjects*sizeof(*work));
	for (i = 0; i < NAREAL; i++)
	    areal[i] = cpl_table_get_data_float(tab,col_areals[i]);

	/* Do the seeing calculation */

	seeing(ap,nobjects,ellipt,pkht,areal,work,&fwhm);
    } else 
	fwhm = 0.0;
    ap->fwhm = fwhm;

    /* Get out of here */

    tidy();
    return(VIR_OK);
}

static void tidy (void) {

    /* Free up workspace */

    freespace(work);
}

/**@}*/

/*

$Log: create_table.c,v $
Revision 1.5  2012/01/15 17:40:09  jim
Minor modifications to take into accout the changes in cpl API for v6

Revision 1.4  2010/09/09 12:09:57  jim
Added docs

Revision 1.3  2007/03/01 12:38:26  jim
Small modifications after a bit of code checking

Revision 1.2  2006/11/10 09:25:39  jim
Modifed tabinit to only start with a single row

Revision 1.1  2005/09/13 13:25:27  jim
Initial entry after modifications to make cpl compliant


*/
